######################
# fickle: Responsively updating variables
#
# Define & recompute shared variables based upon viewport changes, such as a 
# resize of a window. 

# To begin, call fickle.init()

# If you want to have more than just window_width and window_height defined, 
# pass a function (vars) into init and return the custom variables you 
# want. vars are the default responsive variables (at this time, window 
# width and height). For example: 

# fickle.init (vars) -> 
#   single_col: vars.window_width < 500
#   gutter: if vars.window_width > 1000 then 80 else 10

# If you're using include statements, it might look like: 

# include '/code/considerit/fickle.coffee', -> 
#  fickle.init (vars) -> 
#    gutter: 20 
#    one_col: vars.window_width < 675

# All variables are made available on fickle. E.g. fickle.window_width

# If any of these variables ever changes, any component that has used them will 
# be re-rendered. 

# Beneath the surface, all variables are at fetch("fickle_vars"), and a getter 
# is used to facilitate the direct variable access while subscribing callers 
# to changes. 


fickle_events = ['resize']
  # Events that should trigger a recomputation of variables. 
  # Right now, only responding to window resize events. 

window.fickle = 

  init: (client_vars) -> 
    fickle.client_vars = client_vars

    for ev in fickle_events
      window.addEventListener ev, fickle.recompute_vars

    # Initialize the variables
    vars = fickle.recompute_vars()

    # Convenience method for programmers to access variables.
    for lvar in Object.keys(vars)
      do (lvar) ->
        Object.defineProperty fickle, lvar,
          get: -> 
            fetch('fickle_vars')[lvar]


  recompute_vars: -> 
    vars = fetch('fickle_vars')

    base_vars = 
      window_width: window.innerWidth
      window_height: window.innerHeight

    # Compute the custom variables that the programmer wants defined
    if fickle.client_vars
      client_vars = fickle.client_vars base_vars
    else 
      client_vars = {}

    new_vals = extend client_vars, base_vars 


    # only update if we have a change
    changed = false
    for own k,v of new_vals
      if vars[k] != v
        changed = true
        vars[k] = v

    save(vars) if changed
    vars

